from ctc import spec


contract_examples: dict[str, spec.Address] = {
    'PoolDirectory': '0x835482fe0532f169024d5e9410199369aad5c77e',
    'Token': '0xfd3300a9a74b3250f1b2abc12b47611171910b07',
    'Comptroller': '0xc54172e34046c1653d1920d40333dd358c7a1af4',
    'InterestRateModel': '0x075538650a9c69ac8019507a7dd1bd879b12c1d7',
    'Oracle': '0x4d10bc156fbad2474a94f792fe0d6c3261469cdd',
}


pool_directory_abis = {
    'AdminChanged': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'previousAdmin',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'newAdmin',
                'type': 'address',
            },
        ],
        'name': 'AdminChanged',
        'type': 'event',
    },
    'Upgraded': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'implementation',
                'type': 'address',
            }
        ],
        'name': 'Upgraded',
        'type': 'event',
    },
    'admin': {
        'inputs': [],
        'name': 'admin',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'changeAdmin': {
        'inputs': [
            {'internalType': 'address', 'name': 'newAdmin', 'type': 'address'}
        ],
        'name': 'changeAdmin',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'implementation': {
        'inputs': [],
        'name': 'implementation',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'upgradeTo': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'newImplementation',
                'type': 'address',
            }
        ],
        'name': 'upgradeTo',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'upgradeToAndCall': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'newImplementation',
                'type': 'address',
            },
            {'internalType': 'bytes', 'name': 'data', 'type': 'bytes'},
        ],
        'name': 'upgradeToAndCall',
        'outputs': [],
        'stateMutability': 'payable',
        'type': 'function',
    },
    'AdminWhitelistUpdated': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address[]',
                'name': 'admins',
                'type': 'address[]',
            },
            {
                'indexed': False,
                'internalType': 'bool',
                'name': 'status',
                'type': 'bool',
            },
        ],
        'name': 'AdminWhitelistUpdated',
        'type': 'event',
    },
    'OwnershipTransferred': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'previousOwner',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'newOwner',
                'type': 'address',
            },
        ],
        'name': 'OwnershipTransferred',
        'type': 'event',
    },
    'PoolRegistered': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'index',
                'type': 'uint256',
            },
            {
                'components': [
                    {
                        'internalType': 'string',
                        'name': 'name',
                        'type': 'string',
                    },
                    {
                        'internalType': 'address',
                        'name': 'creator',
                        'type': 'address',
                    },
                    {
                        'internalType': 'address',
                        'name': 'comptroller',
                        'type': 'address',
                    },
                    {
                        'internalType': 'uint256',
                        'name': 'blockPosted',
                        'type': 'uint256',
                    },
                    {
                        'internalType': 'uint256',
                        'name': 'timestampPosted',
                        'type': 'uint256',
                    },
                ],
                'indexed': False,
                'internalType': 'struct FusePoolDirectory.FusePool',
                'name': 'pool',
                'type': 'tuple',
            },
        ],
        'name': 'PoolRegistered',
        'type': 'event',
    },
    '_editAdminWhitelist': {
        'inputs': [
            {
                'internalType': 'address[]',
                'name': 'admins',
                'type': 'address[]',
            },
            {'internalType': 'bool', 'name': 'status', 'type': 'bool'},
        ],
        'name': '_editAdminWhitelist',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_editDeployerWhitelist': {
        'inputs': [
            {
                'internalType': 'address[]',
                'name': 'deployers',
                'type': 'address[]',
            },
            {'internalType': 'bool', 'name': 'status', 'type': 'bool'},
        ],
        'name': '_editDeployerWhitelist',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setDeployerWhitelistEnforcement': {
        'inputs': [{'internalType': 'bool', 'name': 'enforce', 'type': 'bool'}],
        'name': '_setDeployerWhitelistEnforcement',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'adminWhitelist': {
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'adminWhitelist',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'bookmarkPool': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'comptroller',
                'type': 'address',
            }
        ],
        'name': 'bookmarkPool',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'deployPool': {
        'inputs': [
            {'internalType': 'string', 'name': 'name', 'type': 'string'},
            {
                'internalType': 'address',
                'name': 'implementation',
                'type': 'address',
            },
            {
                'internalType': 'bool',
                'name': 'enforceWhitelist',
                'type': 'bool',
            },
            {
                'internalType': 'uint256',
                'name': 'closeFactor',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'liquidationIncentive',
                'type': 'uint256',
            },
            {
                'internalType': 'address',
                'name': 'priceOracle',
                'type': 'address',
            },
        ],
        'name': 'deployPool',
        'outputs': [
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
            {'internalType': 'address', 'name': '', 'type': 'address'},
        ],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'deployerWhitelist': {
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'deployerWhitelist',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'enforceDeployerWhitelist': {
        'inputs': [],
        'name': 'enforceDeployerWhitelist',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getAllPools': {
        'inputs': [],
        'name': 'getAllPools',
        'outputs': [
            {
                'components': [
                    {
                        'internalType': 'string',
                        'name': 'name',
                        'type': 'string',
                    },
                    {
                        'internalType': 'address',
                        'name': 'creator',
                        'type': 'address',
                    },
                    {
                        'internalType': 'address',
                        'name': 'comptroller',
                        'type': 'address',
                    },
                    {
                        'internalType': 'uint256',
                        'name': 'blockPosted',
                        'type': 'uint256',
                    },
                    {
                        'internalType': 'uint256',
                        'name': 'timestampPosted',
                        'type': 'uint256',
                    },
                ],
                'internalType': 'struct FusePoolDirectory.FusePool[]',
                'name': '',
                'type': 'tuple[]',
            }
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getBookmarks': {
        'inputs': [
            {'internalType': 'address', 'name': 'account', 'type': 'address'}
        ],
        'name': 'getBookmarks',
        'outputs': [
            {'internalType': 'address[]', 'name': '', 'type': 'address[]'}
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getPoolsByAccount': {
        'inputs': [
            {'internalType': 'address', 'name': 'account', 'type': 'address'}
        ],
        'name': 'getPoolsByAccount',
        'outputs': [
            {'internalType': 'uint256[]', 'name': '', 'type': 'uint256[]'},
            {
                'components': [
                    {
                        'internalType': 'string',
                        'name': 'name',
                        'type': 'string',
                    },
                    {
                        'internalType': 'address',
                        'name': 'creator',
                        'type': 'address',
                    },
                    {
                        'internalType': 'address',
                        'name': 'comptroller',
                        'type': 'address',
                    },
                    {
                        'internalType': 'uint256',
                        'name': 'blockPosted',
                        'type': 'uint256',
                    },
                    {
                        'internalType': 'uint256',
                        'name': 'timestampPosted',
                        'type': 'uint256',
                    },
                ],
                'internalType': 'struct FusePoolDirectory.FusePool[]',
                'name': '',
                'type': 'tuple[]',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getPublicPools': {
        'inputs': [],
        'name': 'getPublicPools',
        'outputs': [
            {'internalType': 'uint256[]', 'name': '', 'type': 'uint256[]'},
            {
                'components': [
                    {
                        'internalType': 'string',
                        'name': 'name',
                        'type': 'string',
                    },
                    {
                        'internalType': 'address',
                        'name': 'creator',
                        'type': 'address',
                    },
                    {
                        'internalType': 'address',
                        'name': 'comptroller',
                        'type': 'address',
                    },
                    {
                        'internalType': 'uint256',
                        'name': 'blockPosted',
                        'type': 'uint256',
                    },
                    {
                        'internalType': 'uint256',
                        'name': 'timestampPosted',
                        'type': 'uint256',
                    },
                ],
                'internalType': 'struct FusePoolDirectory.FusePool[]',
                'name': '',
                'type': 'tuple[]',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getPublicPoolsByVerification': {
        'inputs': [
            {'internalType': 'bool', 'name': 'whitelistedAdmin', 'type': 'bool'}
        ],
        'name': 'getPublicPoolsByVerification',
        'outputs': [
            {'internalType': 'uint256[]', 'name': '', 'type': 'uint256[]'},
            {
                'components': [
                    {
                        'internalType': 'string',
                        'name': 'name',
                        'type': 'string',
                    },
                    {
                        'internalType': 'address',
                        'name': 'creator',
                        'type': 'address',
                    },
                    {
                        'internalType': 'address',
                        'name': 'comptroller',
                        'type': 'address',
                    },
                    {
                        'internalType': 'uint256',
                        'name': 'blockPosted',
                        'type': 'uint256',
                    },
                    {
                        'internalType': 'uint256',
                        'name': 'timestampPosted',
                        'type': 'uint256',
                    },
                ],
                'internalType': 'struct FusePoolDirectory.FusePool[]',
                'name': '',
                'type': 'tuple[]',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'initialize': {
        'inputs': [
            {
                'internalType': 'bool',
                'name': '_enforceDeployerWhitelist',
                'type': 'bool',
            },
            {
                'internalType': 'address[]',
                'name': '_deployerWhitelist',
                'type': 'address[]',
            },
        ],
        'name': 'initialize',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'owner': {
        'inputs': [],
        'name': 'owner',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'poolExists': {
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'poolExists',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'pools': {
        'inputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'name': 'pools',
        'outputs': [
            {'internalType': 'string', 'name': 'name', 'type': 'string'},
            {'internalType': 'address', 'name': 'creator', 'type': 'address'},
            {
                'internalType': 'address',
                'name': 'comptroller',
                'type': 'address',
            },
            {
                'internalType': 'uint256',
                'name': 'blockPosted',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'timestampPosted',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'renounceOwnership': {
        'inputs': [],
        'name': 'renounceOwnership',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'setPoolName': {
        'inputs': [
            {'internalType': 'uint256', 'name': 'index', 'type': 'uint256'},
            {'internalType': 'string', 'name': 'name', 'type': 'string'},
        ],
        'name': 'setPoolName',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'transferOwnership': {
        'inputs': [
            {'internalType': 'address', 'name': 'newOwner', 'type': 'address'}
        ],
        'name': 'transferOwnership',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
}


comptroller_abis = {
    'ActionPaused': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'string',
                'name': 'action',
                'type': 'string',
            },
            {
                'indexed': False,
                'internalType': 'bool',
                'name': 'pauseState',
                'type': 'bool',
            },
        ],
        'name': 'ActionPaused',
        'type': 'event',
    },
    'AddedRewardsDistributor': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'rewardsDistributor',
                'type': 'address',
            }
        ],
        'name': 'AddedRewardsDistributor',
        'type': 'event',
    },
    'AutoImplementationsToggled': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'bool',
                'name': 'enabled',
                'type': 'bool',
            }
        ],
        'name': 'AutoImplementationsToggled',
        'type': 'event',
    },
    'Failure': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'error',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'info',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'detail',
                'type': 'uint256',
            },
        ],
        'name': 'Failure',
        'type': 'event',
    },
    'MarketEntered': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'account',
                'type': 'address',
            },
        ],
        'name': 'MarketEntered',
        'type': 'event',
    },
    'MarketExited': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'account',
                'type': 'address',
            },
        ],
        'name': 'MarketExited',
        'type': 'event',
    },
    'MarketListed': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            }
        ],
        'name': 'MarketListed',
        'type': 'event',
    },
    'MarketUnlisted': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            }
        ],
        'name': 'MarketUnlisted',
        'type': 'event',
    },
    'NewBorrowCap': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'newBorrowCap',
                'type': 'uint256',
            },
        ],
        'name': 'NewBorrowCap',
        'type': 'event',
    },
    'NewBorrowCapGuardian': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'oldBorrowCapGuardian',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'newBorrowCapGuardian',
                'type': 'address',
            },
        ],
        'name': 'NewBorrowCapGuardian',
        'type': 'event',
    },
    'NewCloseFactor': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'oldCloseFactorMantissa',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'newCloseFactorMantissa',
                'type': 'uint256',
            },
        ],
        'name': 'NewCloseFactor',
        'type': 'event',
    },
    'NewCollateralFactor': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'oldCollateralFactorMantissa',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'newCollateralFactorMantissa',
                'type': 'uint256',
            },
        ],
        'name': 'NewCollateralFactor',
        'type': 'event',
    },
    'NewLiquidationIncentive': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'oldLiquidationIncentiveMantissa',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'newLiquidationIncentiveMantissa',
                'type': 'uint256',
            },
        ],
        'name': 'NewLiquidationIncentive',
        'type': 'event',
    },
    'NewPauseGuardian': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'oldPauseGuardian',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'newPauseGuardian',
                'type': 'address',
            },
        ],
        'name': 'NewPauseGuardian',
        'type': 'event',
    },
    'NewPriceOracle': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'contract PriceOracle',
                'name': 'oldPriceOracle',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'contract PriceOracle',
                'name': 'newPriceOracle',
                'type': 'address',
            },
        ],
        'name': 'NewPriceOracle',
        'type': 'event',
    },
    'NewSupplyCap': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'newSupplyCap',
                'type': 'uint256',
            },
        ],
        'name': 'NewSupplyCap',
        'type': 'event',
    },
    'WhitelistEnforcementChanged': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'bool',
                'name': 'enforce',
                'type': 'bool',
            }
        ],
        'name': 'WhitelistEnforcementChanged',
        'type': 'event',
    },
    '_addRewardsDistributor': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'distributor',
                'type': 'address',
            }
        ],
        'name': '_addRewardsDistributor',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_afterNonReentrant': {
        'constant': False,
        'inputs': [],
        'name': '_afterNonReentrant',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_become': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'contract Unitroller',
                'name': 'unitroller',
                'type': 'address',
            }
        ],
        'name': '_become',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_becomeImplementation': {
        'constant': False,
        'inputs': [],
        'name': '_becomeImplementation',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_beforeNonReentrant': {
        'constant': False,
        'inputs': [],
        'name': '_beforeNonReentrant',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_borrowGuardianPaused': {
        'constant': True,
        'inputs': [],
        'name': '_borrowGuardianPaused',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    '_deployMarket': {
        'constant': False,
        'inputs': [
            {'internalType': 'bool', 'name': 'isCEther', 'type': 'bool'},
            {
                'internalType': 'bytes',
                'name': 'constructorData',
                'type': 'bytes',
            },
            {
                'internalType': 'uint256',
                'name': 'collateralFactorMantissa',
                'type': 'uint256',
            },
        ],
        'name': '_deployMarket',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_mintGuardianPaused': {
        'constant': True,
        'inputs': [],
        'name': '_mintGuardianPaused',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    '_setBorrowCapGuardian': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'newBorrowCapGuardian',
                'type': 'address',
            }
        ],
        'name': '_setBorrowCapGuardian',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setBorrowPaused': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            },
            {'internalType': 'bool', 'name': 'state', 'type': 'bool'},
        ],
        'name': '_setBorrowPaused',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setCloseFactor': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'newCloseFactorMantissa',
                'type': 'uint256',
            }
        ],
        'name': '_setCloseFactor',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setCollateralFactor': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            },
            {
                'internalType': 'uint256',
                'name': 'newCollateralFactorMantissa',
                'type': 'uint256',
            },
        ],
        'name': '_setCollateralFactor',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setLiquidationIncentive': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'newLiquidationIncentiveMantissa',
                'type': 'uint256',
            }
        ],
        'name': '_setLiquidationIncentive',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setMarketBorrowCaps': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'contract CToken[]',
                'name': 'cTokens',
                'type': 'address[]',
            },
            {
                'internalType': 'uint256[]',
                'name': 'newBorrowCaps',
                'type': 'uint256[]',
            },
        ],
        'name': '_setMarketBorrowCaps',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setMarketSupplyCaps': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'contract CToken[]',
                'name': 'cTokens',
                'type': 'address[]',
            },
            {
                'internalType': 'uint256[]',
                'name': 'newSupplyCaps',
                'type': 'uint256[]',
            },
        ],
        'name': '_setMarketSupplyCaps',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setMintPaused': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            },
            {'internalType': 'bool', 'name': 'state', 'type': 'bool'},
        ],
        'name': '_setMintPaused',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setPauseGuardian': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'newPauseGuardian',
                'type': 'address',
            }
        ],
        'name': '_setPauseGuardian',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setPriceOracle': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'contract PriceOracle',
                'name': 'newOracle',
                'type': 'address',
            }
        ],
        'name': '_setPriceOracle',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setSeizePaused': {
        'constant': False,
        'inputs': [{'internalType': 'bool', 'name': 'state', 'type': 'bool'}],
        'name': '_setSeizePaused',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setTransferPaused': {
        'constant': False,
        'inputs': [{'internalType': 'bool', 'name': 'state', 'type': 'bool'}],
        'name': '_setTransferPaused',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setWhitelistEnforcement': {
        'constant': False,
        'inputs': [{'internalType': 'bool', 'name': 'enforce', 'type': 'bool'}],
        'name': '_setWhitelistEnforcement',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setWhitelistStatuses': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address[]',
                'name': 'suppliers',
                'type': 'address[]',
            },
            {'internalType': 'bool[]', 'name': 'statuses', 'type': 'bool[]'},
        ],
        'name': '_setWhitelistStatuses',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_toggleAutoImplementations': {
        'constant': False,
        'inputs': [{'internalType': 'bool', 'name': 'enabled', 'type': 'bool'}],
        'name': '_toggleAutoImplementations',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_unsupportMarket': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            }
        ],
        'name': '_unsupportMarket',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'accountAssets': {
        'constant': True,
        'inputs': [
            {'internalType': 'address', 'name': '', 'type': 'address'},
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
        ],
        'name': 'accountAssets',
        'outputs': [
            {'internalType': 'contract CToken', 'name': '', 'type': 'address'}
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'admin': {
        'constant': True,
        'inputs': [],
        'name': 'admin',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'adminHasRights': {
        'constant': True,
        'inputs': [],
        'name': 'adminHasRights',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'allBorrowers': {
        'constant': True,
        'inputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'name': 'allBorrowers',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'allMarkets': {
        'constant': True,
        'inputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'name': 'allMarkets',
        'outputs': [
            {'internalType': 'contract CToken', 'name': '', 'type': 'address'}
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'autoImplementation': {
        'constant': True,
        'inputs': [],
        'name': 'autoImplementation',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'borrowAllowed': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'borrowAmount',
                'type': 'uint256',
            },
        ],
        'name': 'borrowAllowed',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'borrowCapGuardian': {
        'constant': True,
        'inputs': [],
        'name': 'borrowCapGuardian',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'borrowCaps': {
        'constant': True,
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'borrowCaps',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'borrowGuardianPaused': {
        'constant': True,
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'borrowGuardianPaused',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'borrowVerify': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'borrowAmount',
                'type': 'uint256',
            },
        ],
        'name': 'borrowVerify',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'borrowWithinLimits': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'accountBorrowsNew',
                'type': 'uint256',
            },
        ],
        'name': 'borrowWithinLimits',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'cTokensByUnderlying': {
        'constant': True,
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'cTokensByUnderlying',
        'outputs': [
            {'internalType': 'contract CToken', 'name': '', 'type': 'address'}
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'checkMembership': {
        'constant': True,
        'inputs': [
            {'internalType': 'address', 'name': 'account', 'type': 'address'},
            {
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            },
        ],
        'name': 'checkMembership',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'closeFactorMantissa': {
        'constant': True,
        'inputs': [],
        'name': 'closeFactorMantissa',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'comptrollerImplementation': {
        'constant': True,
        'inputs': [],
        'name': 'comptrollerImplementation',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'enforceWhitelist': {
        'constant': True,
        'inputs': [],
        'name': 'enforceWhitelist',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'enterMarkets': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address[]',
                'name': 'cTokens',
                'type': 'address[]',
            }
        ],
        'name': 'enterMarkets',
        'outputs': [
            {'internalType': 'uint256[]', 'name': '', 'type': 'uint256[]'}
        ],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'exitMarket': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'cTokenAddress',
                'type': 'address',
            }
        ],
        'name': 'exitMarket',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'fuseAdminHasRights': {
        'constant': True,
        'inputs': [],
        'name': 'fuseAdminHasRights',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getAccountLiquidity': {
        'constant': True,
        'inputs': [
            {'internalType': 'address', 'name': 'account', 'type': 'address'}
        ],
        'name': 'getAccountLiquidity',
        'outputs': [
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getAllBorrowers': {
        'constant': True,
        'inputs': [],
        'name': 'getAllBorrowers',
        'outputs': [
            {'internalType': 'address[]', 'name': '', 'type': 'address[]'}
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getAllMarkets': {
        'constant': True,
        'inputs': [],
        'name': 'getAllMarkets',
        'outputs': [
            {
                'internalType': 'contract CToken[]',
                'name': '',
                'type': 'address[]',
            }
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getAssetsIn': {
        'constant': True,
        'inputs': [
            {'internalType': 'address', 'name': 'account', 'type': 'address'}
        ],
        'name': 'getAssetsIn',
        'outputs': [
            {
                'internalType': 'contract CToken[]',
                'name': '',
                'type': 'address[]',
            }
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getHypotheticalAccountLiquidity': {
        'constant': True,
        'inputs': [
            {'internalType': 'address', 'name': 'account', 'type': 'address'},
            {
                'internalType': 'address',
                'name': 'cTokenModify',
                'type': 'address',
            },
            {
                'internalType': 'uint256',
                'name': 'redeemTokens',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'borrowAmount',
                'type': 'uint256',
            },
        ],
        'name': 'getHypotheticalAccountLiquidity',
        'outputs': [
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getRewardsDistributors': {
        'constant': True,
        'inputs': [],
        'name': 'getRewardsDistributors',
        'outputs': [
            {'internalType': 'address[]', 'name': '', 'type': 'address[]'}
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getWhitelist': {
        'constant': True,
        'inputs': [],
        'name': 'getWhitelist',
        'outputs': [
            {'internalType': 'address[]', 'name': '', 'type': 'address[]'}
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'isComptroller': {
        'constant': True,
        'inputs': [],
        'name': 'isComptroller',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'isDeprecated': {
        'constant': True,
        'inputs': [
            {
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            }
        ],
        'name': 'isDeprecated',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'liquidateBorrowAllowed': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'cTokenBorrowed',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'cTokenCollateral',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'liquidator',
                'type': 'address',
            },
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'repayAmount',
                'type': 'uint256',
            },
        ],
        'name': 'liquidateBorrowAllowed',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'liquidateBorrowVerify': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'cTokenBorrowed',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'cTokenCollateral',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'liquidator',
                'type': 'address',
            },
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'actualRepayAmount',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'seizeTokens',
                'type': 'uint256',
            },
        ],
        'name': 'liquidateBorrowVerify',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'liquidateCalculateSeizeTokens': {
        'constant': True,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'cTokenBorrowed',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'cTokenCollateral',
                'type': 'address',
            },
            {
                'internalType': 'uint256',
                'name': 'actualRepayAmount',
                'type': 'uint256',
            },
        ],
        'name': 'liquidateCalculateSeizeTokens',
        'outputs': [
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'liquidationIncentiveMantissa': {
        'constant': True,
        'inputs': [],
        'name': 'liquidationIncentiveMantissa',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'markets': {
        'constant': True,
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'markets',
        'outputs': [
            {'internalType': 'bool', 'name': 'isListed', 'type': 'bool'},
            {
                'internalType': 'uint256',
                'name': 'collateralFactorMantissa',
                'type': 'uint256',
            },
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'mintAllowed': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {'internalType': 'address', 'name': 'minter', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'mintAmount',
                'type': 'uint256',
            },
        ],
        'name': 'mintAllowed',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'mintGuardianPaused': {
        'constant': True,
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'mintGuardianPaused',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'mintVerify': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {'internalType': 'address', 'name': 'minter', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'actualMintAmount',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'mintTokens',
                'type': 'uint256',
            },
        ],
        'name': 'mintVerify',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'mintWithinLimits': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'exchangeRateMantissa',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'accountTokens',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'mintAmount',
                'type': 'uint256',
            },
        ],
        'name': 'mintWithinLimits',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'oracle': {
        'constant': True,
        'inputs': [],
        'name': 'oracle',
        'outputs': [
            {
                'internalType': 'contract PriceOracle',
                'name': '',
                'type': 'address',
            }
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'pauseGuardian': {
        'constant': True,
        'inputs': [],
        'name': 'pauseGuardian',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'pendingAdmin': {
        'constant': True,
        'inputs': [],
        'name': 'pendingAdmin',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'pendingComptrollerImplementation': {
        'constant': True,
        'inputs': [],
        'name': 'pendingComptrollerImplementation',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'redeemAllowed': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {'internalType': 'address', 'name': 'redeemer', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'redeemTokens',
                'type': 'uint256',
            },
        ],
        'name': 'redeemAllowed',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'redeemVerify': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {'internalType': 'address', 'name': 'redeemer', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'redeemAmount',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'redeemTokens',
                'type': 'uint256',
            },
        ],
        'name': 'redeemVerify',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'repayBorrowAllowed': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {'internalType': 'address', 'name': 'payer', 'type': 'address'},
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'repayAmount',
                'type': 'uint256',
            },
        ],
        'name': 'repayBorrowAllowed',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'repayBorrowVerify': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {'internalType': 'address', 'name': 'payer', 'type': 'address'},
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'actualRepayAmount',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'borrowerIndex',
                'type': 'uint256',
            },
        ],
        'name': 'repayBorrowVerify',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'rewardsDistributors': {
        'constant': True,
        'inputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'name': 'rewardsDistributors',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'seizeAllowed': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'cTokenCollateral',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'cTokenBorrowed',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'liquidator',
                'type': 'address',
            },
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'seizeTokens',
                'type': 'uint256',
            },
        ],
        'name': 'seizeAllowed',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'seizeGuardianPaused': {
        'constant': True,
        'inputs': [],
        'name': 'seizeGuardianPaused',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'seizeVerify': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'cTokenCollateral',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'cTokenBorrowed',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'liquidator',
                'type': 'address',
            },
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'seizeTokens',
                'type': 'uint256',
            },
        ],
        'name': 'seizeVerify',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'suppliers': {
        'constant': True,
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'suppliers',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'supplyCaps': {
        'constant': True,
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'supplyCaps',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'transferAllowed': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {'internalType': 'address', 'name': 'src', 'type': 'address'},
            {'internalType': 'address', 'name': 'dst', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'transferTokens',
                'type': 'uint256',
            },
        ],
        'name': 'transferAllowed',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'transferGuardianPaused': {
        'constant': True,
        'inputs': [],
        'name': 'transferGuardianPaused',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'transferVerify': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'cToken', 'type': 'address'},
            {'internalType': 'address', 'name': 'src', 'type': 'address'},
            {'internalType': 'address', 'name': 'dst', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'transferTokens',
                'type': 'uint256',
            },
        ],
        'name': 'transferVerify',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'whitelist': {
        'constant': True,
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'whitelist',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'whitelistArray': {
        'constant': True,
        'inputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'name': 'whitelistArray',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
}

ctoken_function_abis = {
    '_setImplementation': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'implementation_',
                'type': 'address',
            },
            {'internalType': 'bool', 'name': 'allowResign', 'type': 'bool'},
            {
                'internalType': 'bytes',
                'name': 'becomeImplementationData',
                'type': 'bytes',
            },
        ],
        'name': '_setImplementation',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'admin': {
        'constant': True,
        'inputs': [],
        'name': 'admin',
        'outputs': [
            {'internalType': 'address payable', 'name': '', 'type': 'address'}
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'adminHasRights': {
        'constant': True,
        'inputs': [],
        'name': 'adminHasRights',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'delegateToImplementation': {
        'constant': False,
        'inputs': [{'internalType': 'bytes', 'name': 'data', 'type': 'bytes'}],
        'name': 'delegateToImplementation',
        'outputs': [{'internalType': 'bytes', 'name': '', 'type': 'bytes'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'fuseAdminHasRights': {
        'constant': True,
        'inputs': [],
        'name': 'fuseAdminHasRights',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'implementation': {
        'constant': True,
        'inputs': [],
        'name': 'implementation',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    '_acceptAdmin': {
        'constant': False,
        'inputs': [],
        'name': '_acceptAdmin',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_addReserves': {
        'constant': False,
        'inputs': [
            {'internalType': 'uint256', 'name': 'addAmount', 'type': 'uint256'}
        ],
        'name': '_addReserves',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_becomeImplementation': {
        'constant': False,
        'inputs': [{'internalType': 'bytes', 'name': 'data', 'type': 'bytes'}],
        'name': '_becomeImplementation',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_reduceReserves': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'reduceAmount',
                'type': 'uint256',
            }
        ],
        'name': '_reduceReserves',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_renounceAdminRights': {
        'constant': False,
        'inputs': [],
        'name': '_renounceAdminRights',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_renounceFuseAdminRights': {
        'constant': False,
        'inputs': [],
        'name': '_renounceFuseAdminRights',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_resignImplementation': {
        'constant': False,
        'inputs': [],
        'name': '_resignImplementation',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setAdminFee': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'newAdminFeeMantissa',
                'type': 'uint256',
            }
        ],
        'name': '_setAdminFee',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setComptroller': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'contract ComptrollerInterface',
                'name': 'newComptroller',
                'type': 'address',
            }
        ],
        'name': '_setComptroller',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setFuseFee': {
        'constant': False,
        'inputs': [],
        'name': '_setFuseFee',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setInterestRateModel': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'contract InterestRateModel',
                'name': 'newInterestRateModel',
                'type': 'address',
            }
        ],
        'name': '_setInterestRateModel',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setPendingAdmin': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address payable',
                'name': 'newPendingAdmin',
                'type': 'address',
            }
        ],
        'name': '_setPendingAdmin',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_setReserveFactor': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'newReserveFactorMantissa',
                'type': 'uint256',
            }
        ],
        'name': '_setReserveFactor',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_withdrawAdminFees': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'withdrawAmount',
                'type': 'uint256',
            }
        ],
        'name': '_withdrawAdminFees',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    '_withdrawFuseFees': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'withdrawAmount',
                'type': 'uint256',
            }
        ],
        'name': '_withdrawFuseFees',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'accrualBlockNumber': {
        'constant': True,
        'inputs': [],
        'name': 'accrualBlockNumber',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'accrueInterest': {
        'constant': False,
        'inputs': [],
        'name': 'accrueInterest',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'adminFeeMantissa': {
        'constant': True,
        'inputs': [],
        'name': 'adminFeeMantissa',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'allowance': {
        'constant': True,
        'inputs': [
            {'internalType': 'address', 'name': 'owner', 'type': 'address'},
            {'internalType': 'address', 'name': 'spender', 'type': 'address'},
        ],
        'name': 'allowance',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'approve': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'spender', 'type': 'address'},
            {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'},
        ],
        'name': 'approve',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'balanceOf': {
        'constant': True,
        'inputs': [
            {'internalType': 'address', 'name': 'owner', 'type': 'address'}
        ],
        'name': 'balanceOf',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'balanceOfUnderlying': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'owner', 'type': 'address'}
        ],
        'name': 'balanceOfUnderlying',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'borrow': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'borrowAmount',
                'type': 'uint256',
            }
        ],
        'name': 'borrow',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'borrowBalanceCurrent': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'account', 'type': 'address'}
        ],
        'name': 'borrowBalanceCurrent',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'borrowBalanceStored': {
        'constant': True,
        'inputs': [
            {'internalType': 'address', 'name': 'account', 'type': 'address'}
        ],
        'name': 'borrowBalanceStored',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'borrowIndex': {
        'constant': True,
        'inputs': [],
        'name': 'borrowIndex',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'borrowRatePerBlock': {
        'constant': True,
        'inputs': [],
        'name': 'borrowRatePerBlock',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'comptroller': {
        'constant': True,
        'inputs': [],
        'name': 'comptroller',
        'outputs': [
            {
                'internalType': 'contract ComptrollerInterface',
                'name': '',
                'type': 'address',
            }
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'decimals': {
        'constant': True,
        'inputs': [],
        'name': 'decimals',
        'outputs': [{'internalType': 'uint8', 'name': '', 'type': 'uint8'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'exchangeRateCurrent': {
        'constant': False,
        'inputs': [],
        'name': 'exchangeRateCurrent',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'exchangeRateStored': {
        'constant': True,
        'inputs': [],
        'name': 'exchangeRateStored',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'fuseFeeMantissa': {
        'constant': True,
        'inputs': [],
        'name': 'fuseFeeMantissa',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getAccountSnapshot': {
        'constant': True,
        'inputs': [
            {'internalType': 'address', 'name': 'account', 'type': 'address'}
        ],
        'name': 'getAccountSnapshot',
        'outputs': [
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': '', 'type': 'uint256'},
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getCash': {
        'constant': True,
        'inputs': [],
        'name': 'getCash',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'initialize': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'underlying_',
                'type': 'address',
            },
            {
                'internalType': 'contract ComptrollerInterface',
                'name': 'comptroller_',
                'type': 'address',
            },
            {
                'internalType': 'contract InterestRateModel',
                'name': 'interestRateModel_',
                'type': 'address',
            },
            {
                'internalType': 'uint256',
                'name': 'initialExchangeRateMantissa_',
                'type': 'uint256',
            },
            {'internalType': 'string', 'name': 'name_', 'type': 'string'},
            {'internalType': 'string', 'name': 'symbol_', 'type': 'string'},
            {'internalType': 'uint8', 'name': 'decimals_', 'type': 'uint8'},
            {
                'internalType': 'uint256',
                'name': 'reserveFactorMantissa_',
                'type': 'uint256',
            },
            {
                'internalType': 'uint256',
                'name': 'adminFeeMantissa_',
                'type': 'uint256',
            },
        ],
        'name': 'initialize',
        'outputs': [],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'interestRateModel': {
        'constant': True,
        'inputs': [],
        'name': 'interestRateModel',
        'outputs': [
            {
                'internalType': 'contract InterestRateModel',
                'name': '',
                'type': 'address',
            }
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'isCEther': {
        'constant': True,
        'inputs': [],
        'name': 'isCEther',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'isCToken': {
        'constant': True,
        'inputs': [],
        'name': 'isCToken',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'liquidateBorrow': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'repayAmount',
                'type': 'uint256',
            },
            {
                'internalType': 'contract CTokenInterface',
                'name': 'cTokenCollateral',
                'type': 'address',
            },
        ],
        'name': 'liquidateBorrow',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'mint': {
        'constant': False,
        'inputs': [
            {'internalType': 'uint256', 'name': 'mintAmount', 'type': 'uint256'}
        ],
        'name': 'mint',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'name': {
        'constant': True,
        'inputs': [],
        'name': 'name',
        'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'pendingAdmin': {
        'constant': True,
        'inputs': [],
        'name': 'pendingAdmin',
        'outputs': [
            {'internalType': 'address payable', 'name': '', 'type': 'address'}
        ],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'redeem': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'redeemTokens',
                'type': 'uint256',
            }
        ],
        'name': 'redeem',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'redeemUnderlying': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'redeemAmount',
                'type': 'uint256',
            }
        ],
        'name': 'redeemUnderlying',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'repayBorrow': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'repayAmount',
                'type': 'uint256',
            }
        ],
        'name': 'repayBorrow',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'repayBorrowBehalf': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'repayAmount',
                'type': 'uint256',
            },
        ],
        'name': 'repayBorrowBehalf',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'reserveFactorMantissa': {
        'constant': True,
        'inputs': [],
        'name': 'reserveFactorMantissa',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'seize': {
        'constant': False,
        'inputs': [
            {
                'internalType': 'address',
                'name': 'liquidator',
                'type': 'address',
            },
            {'internalType': 'address', 'name': 'borrower', 'type': 'address'},
            {
                'internalType': 'uint256',
                'name': 'seizeTokens',
                'type': 'uint256',
            },
        ],
        'name': 'seize',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'supplyRatePerBlock': {
        'constant': True,
        'inputs': [],
        'name': 'supplyRatePerBlock',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'symbol': {
        'constant': True,
        'inputs': [],
        'name': 'symbol',
        'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'totalAdminFees': {
        'constant': True,
        'inputs': [],
        'name': 'totalAdminFees',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'totalBorrows': {
        'constant': True,
        'inputs': [],
        'name': 'totalBorrows',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'totalBorrowsCurrent': {
        'constant': False,
        'inputs': [],
        'name': 'totalBorrowsCurrent',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'totalFuseFees': {
        'constant': True,
        'inputs': [],
        'name': 'totalFuseFees',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'totalReserves': {
        'constant': True,
        'inputs': [],
        'name': 'totalReserves',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'totalSupply': {
        'constant': True,
        'inputs': [],
        'name': 'totalSupply',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'transfer': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'dst', 'type': 'address'},
            {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'},
        ],
        'name': 'transfer',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'transferFrom': {
        'constant': False,
        'inputs': [
            {'internalType': 'address', 'name': 'src', 'type': 'address'},
            {'internalType': 'address', 'name': 'dst', 'type': 'address'},
            {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'},
        ],
        'name': 'transferFrom',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'underlying': {
        'constant': True,
        'inputs': [],
        'name': 'underlying',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
}


interest_rate_model_abis = {
    'baseRatePerBlock': {
        'constant': True,
        'inputs': [],
        'name': 'baseRatePerBlock',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'blocksPerYear': {
        'constant': True,
        'inputs': [],
        'name': 'blocksPerYear',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getBorrowRate': {
        'constant': True,
        'inputs': [
            {'internalType': 'uint256', 'name': 'cash', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': 'borrows', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': 'reserves', 'type': 'uint256'},
        ],
        'name': 'getBorrowRate',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'getSupplyRate': {
        'constant': True,
        'inputs': [
            {'internalType': 'uint256', 'name': 'cash', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': 'borrows', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': 'reserves', 'type': 'uint256'},
            {
                'internalType': 'uint256',
                'name': 'reserveFactorMantissa',
                'type': 'uint256',
            },
        ],
        'name': 'getSupplyRate',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'isInterestRateModel': {
        'constant': True,
        'inputs': [],
        'name': 'isInterestRateModel',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'jumpMultiplierPerBlock': {
        'constant': True,
        'inputs': [],
        'name': 'jumpMultiplierPerBlock',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'kink': {
        'constant': True,
        'inputs': [],
        'name': 'kink',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'multiplierPerBlock': {
        'constant': True,
        'inputs': [],
        'name': 'multiplierPerBlock',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'utilizationRate': {
        'constant': True,
        'inputs': [
            {'internalType': 'uint256', 'name': 'cash', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': 'borrows', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': 'reserves', 'type': 'uint256'},
        ],
        'name': 'utilizationRate',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'pure',
        'type': 'function',
    },
}

oracle_abis = {
    'NewAdmin': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'oldAdmin',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'newAdmin',
                'type': 'address',
            },
        ],
        'name': 'NewAdmin',
        'type': 'event',
    },
    'NewDefaultOracle': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'oldOracle',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'newOracle',
                'type': 'address',
            },
        ],
        'name': 'NewDefaultOracle',
        'type': 'event',
    },
    'NewOracle': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'underlying',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'oldOracle',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'newOracle',
                'type': 'address',
            },
        ],
        'name': 'NewOracle',
        'type': 'event',
    },
    'add': {
        'inputs': [
            {
                'internalType': 'address[]',
                'name': 'underlyings',
                'type': 'address[]',
            },
            {
                'internalType': 'contract PriceOracle[]',
                'name': '_oracles',
                'type': 'address[]',
            },
        ],
        'name': 'add',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'admin': {
        'inputs': [],
        'name': 'admin',
        'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'canAdminOverwrite': {
        'inputs': [],
        'name': 'canAdminOverwrite',
        'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'changeAdmin': {
        'inputs': [
            {'internalType': 'address', 'name': 'newAdmin', 'type': 'address'}
        ],
        'name': 'changeAdmin',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'defaultOracle': {
        'inputs': [],
        'name': 'defaultOracle',
        'outputs': [
            {
                'internalType': 'contract PriceOracle',
                'name': '',
                'type': 'address',
            }
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getUnderlyingPrice': {
        'inputs': [
            {
                'internalType': 'contract CToken',
                'name': 'cToken',
                'type': 'address',
            }
        ],
        'name': 'getUnderlyingPrice',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'initialize': {
        'inputs': [
            {
                'internalType': 'address[]',
                'name': 'underlyings',
                'type': 'address[]',
            },
            {
                'internalType': 'contract PriceOracle[]',
                'name': '_oracles',
                'type': 'address[]',
            },
            {
                'internalType': 'contract PriceOracle',
                'name': '_defaultOracle',
                'type': 'address',
            },
            {'internalType': 'address', 'name': '_admin', 'type': 'address'},
            {
                'internalType': 'bool',
                'name': '_canAdminOverwrite',
                'type': 'bool',
            },
        ],
        'name': 'initialize',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'oracles': {
        'inputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
        'name': 'oracles',
        'outputs': [
            {
                'internalType': 'contract PriceOracle',
                'name': '',
                'type': 'address',
            }
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'price': {
        'inputs': [
            {'internalType': 'address', 'name': 'underlying', 'type': 'address'}
        ],
        'name': 'price',
        'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'setDefaultOracle': {
        'inputs': [
            {
                'internalType': 'contract PriceOracle',
                'name': 'newOracle',
                'type': 'address',
            }
        ],
        'name': 'setDefaultOracle',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
}




erc20_event_abis = {
    'Transfer': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'from',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'to',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amount',
                'type': 'uint256',
            },
        ],
        'name': 'Transfer',
        'type': 'event',
    },
}




from ctc import evm
from ctc import rpc


def get_command_spec():
    return {
        'f': async_block_command,
        'args': [
            {'name': 'block', 'kwargs': {'nargs': '?'}},
            {'name': '--timestamp', 'kwargs': {'type': int}},
            {'name': '--verbose', 'kwargs': {'action': 'store_true'}},
        ],
    }


async def async_block_command(block, timestamp, verbose=False, **kwargs):
    if timestamp is not None and block is None:
        print('searching for block at timestamp:', timestamp)
        block = evm.get_block_of_timestamp(timestamp=timestamp, verbose=False)
        print('found block', block)
        print()

    await evm.async_print_block_summary(block=block)
    await rpc.async_close_http_session()




from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_shh_version(provider=None):
    request = rpc_constructors.construct_shh_version()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_shh_version(
        response=response,
        provider=provider,
    )


async def async_shh_post(topics, payload, priority, ttl, provider=None):
    request = rpc_constructors.construct_ssh_post(
        topics=topics,
        payload=payload,
        priority=priority,
        ttl=ttl,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_shh_post(
        response=response,
        provider=provider,
    )


async def async_shh_new_identity(provider=None):
    request = rpc_constructors.construct_ssh_new_identity()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_ssh_new_identity(
        response=response,
        provider=provider,
    )


async def async_shh_has_identity(address, provider=None):
    request = rpc_constructors.construct_shh_has_identity(address=address)
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_shh_has_identity(
        response=response,
        provider=provider,
    )


async def async_shh_new_group(provider=None):
    request = rpc_constructors.construct_ssh_new_group()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_ssh_new_group(
        response=response,
        provider=provider,
    )


async def async_shh_add_to_group(address, provider=None):
    request = rpc_constructors.construct_ssh_add_to_group(address=address)
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_ssh_add_to_group(
        response=response,
        provider=provider,
    )


async def async_shh_new_filter(receiver, topics, provider=None):
    request = rpc_constructors.construct_shh_new_filter(
        receiver=receiver, topics=topics
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_shh_new_filter(
        response=response,
        provider=provider,
    )


async def async_shh_uninstall_filter(filter_id, provider=None):
    request = rpc_constructors.construct_shh_uninstall_filter(
        filter_id=filter_id
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_shh_uninstall_filter(
        response=response,
        provider=provider,
    )


async def async_shh_get_filter_changes(filter_id, provider=None):
    request = rpc_constructors.construct_ssh_get_filter_changes(
        filter_id=filter_id,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_ssh_get_filter_changes(
        response=response,
        provider=provider,
    )


async def async_shh_get_messages(filter_id, provider=None):
    request = rpc_constructors.construct_ssh_get_messages(filter_id=filter_id)
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_ssh_get_messages(
        response=response,
        provider=provider,
    )




from __future__ import annotations

import typing

from ctc import binary
from ctc import spec
from . import contract_abi_io


async def async_get_function_abi(
    function_name: typing.Optional[str] = None,
    contract_abi: typing.Optional[spec.ContractABI] = None,
    contract_address: typing.Optional[spec.Address] = None,
    n_parameters: typing.Optional[int] = None,
    parameter_types: typing.Optional[list[spec.ABIDatumType]] = None,
    function_selector: typing.Optional[spec.FunctionSelector] = None,
    network: typing.Optional[spec.NetworkReference] = None,
) -> spec.FunctionABI:

    if contract_abi is None:
        contract_abi = await contract_abi_io.async_get_contract_abi(
            contract_address=contract_address,
            network=network,
        )

    return binary.get_function_abi(
        function_name=function_name,
        contract_abi=contract_abi,
        n_parameters=n_parameters,
        parameter_types=parameter_types,
        function_selector=function_selector,
    )




from __future__ import annotations

import asyncio

from ctc import evm
from ctc import rpc
from ctc import spec

from . import uniswap_v2_spec


async def async_get_pool_tokens(
    pool: spec.Address,
    provider: spec.ProviderSpec = None,
) -> tuple[spec.Address, spec.Address]:

    token0 = rpc.async_eth_call(
        function_name='token0', to_address=pool, provider=provider
    )
    token1 = rpc.async_eth_call(
        function_name='token1', to_address=pool, provider=provider
    )

    return await asyncio.gather(token0, token1)


async def async_get_pool_symbols(
    pool: spec.Address = None,
    *,
    x_address: spec.Address = None,
    y_address: spec.Address = None,
    provider: spec.ProviderSpec = None,
):

    if x_address is None or y_address is None:
        if pool is None:
            raise Exception('must specify pool or tokens')
        x_address, y_address = await async_get_pool_tokens(
            pool=pool, provider=provider
        )

    return await evm.async_get_erc20s_symbols(
        tokens=[x_address, y_address],
        provider=provider,
    )


async def async_get_pool_decimals(
    pool: spec.Address = None,
    *,
    x_address: spec.Address = None,
    y_address: spec.Address = None,
    provider: spec.ProviderSpec = None,
):

    if x_address is None or y_address is None:
        if pool is None:
            raise Exception('must specify pool or tokens')
        x_address, y_address = await async_get_pool_tokens(
            pool=pool, provider=provider
        )

    return await evm.async_get_erc20s_decimals(
        tokens=[x_address, y_address],
        provider=provider,
    )


async def async_get_pool_tokens_metadata(
    pool: spec.Address,
    provider: spec.ProviderSpec = None,
) -> uniswap_v2_spec.PoolTokensMetadata:

    x_address, y_address = await async_get_pool_tokens(
        pool=pool,
        provider=provider,
    )

    symbols_coroutine = async_get_pool_symbols(
        x_address=x_address,
        y_address=y_address,
        provider=provider,
    )

    decimals_coroutine = async_get_pool_decimals(
        x_address=x_address,
        y_address=y_address,
        provider=provider,
    )

    symbols, decimals = await asyncio.gather(
        symbols_coroutine, decimals_coroutine
    )
    x_symbol, y_symbol = symbols
    x_decimals, y_decimals = decimals

    return {
        'x_address': x_address,
        'y_address': y_address,
        'x_symbol': x_symbol,
        'y_symbol': y_symbol,
        'x_decimals': x_decimals,
        'y_decimals': y_decimals,
    }




from ctc import evm
from ctc import rpc
from ctc.cli import cli_utils


def get_command_spec():
    return {
        'f': async_transfers_command,
        'args': [
            {'name': 'erc20'},
            {'name': '--blocks', 'kwargs': {'nargs': '+'}},
            {'name': '--output', 'kwargs': {'default': 'stdout'}},
            {'name': '--overwrite', 'kwargs': {'action': 'store_true'}},
        ],
    }


async def async_transfers_command(erc20, blocks, output, overwrite):

    if blocks is not None:
        start_block, end_block = await cli_utils.async_resolve_block_range(
            blocks
        )
    else:
        start_block = None
        end_block = None

    transfers = await evm.async_get_erc20_transfers(
        erc20,
        start_block=start_block,
        end_block=end_block,
    )
    cli_utils.output_data(transfers, output=output, overwrite=overwrite)
    await rpc.async_close_http_session()




from __future__ import annotations

import time
import typing

import toolstr
import tooltime

from ctc import evm
from ctc import spec
from . import analytics_spec


async def async_get_time_data(
    blocks: typing.Sequence[spec.BlockNumberReference] = None,
    timestamps: typing.Sequence[int] = None,
    timescale: typing.Optional[analytics_spec.TimescaleSpec] = None,
    end_time: typing.Optional[analytics_spec.Timestamp] = None,
    window_size: str = None,
    interval_size: str = None,
    provider: spec.ProviderSpec = None,
) -> analytics_spec.TimeData:

    if (
        blocks is None
        or timestamps is None
        or window_size is None
        or interval_size is None
    ):
        if timescale is None:
            raise Exception('must specify timescale or {blocks, timestamps}')
        timescale = resolve_timescale(timescale)
        if timestamps is None:
            if end_time is None:
                end_time = round(time.time())
            timestamps = get_timestamps(timescale=timescale, end_time=end_time)
        if blocks is None:
            blocks = await async_get_timestamps_blocks(
                timestamps=timestamps, provider=provider
            )
        if interval_size is None:
            interval_size = timescale['interval_size']
        if window_size is None:
            window_size = timescale['window_size']

    timestamps = list(timestamps)
    blocks = await evm.async_block_numbers_to_int(
        blocks=blocks, provider=provider
    )

    return {
        'timestamps': timestamps,
        'block_numbers': blocks,
        'n_samples': len(blocks),
        'window_size': window_size,
        'interval_size': interval_size,
        'created_at_timestamp': int(time.time()),
    }


def resolve_timescale(
    timescale: analytics_spec.TimescaleSpec,
) -> analytics_spec.Timescale:
    if (
        isinstance(timescale, dict)
        and set(timescale.keys()) == {'window_size', 'interval_size'}
    ):
        return timescale

    elif isinstance(timescale, str) and timescale.count(',') == 1:
        window_size, interval_size = timescale.split(',')
        window_size = window_size.strip()
        interval_size = interval_size.strip()
        return {'window_size': window_size, 'interval_size': interval_size}

    else:
        raise Exception('could not resolve timescale: ' + str(timescale))


def get_timestamps(
    timescale: analytics_spec.Timescale,
    end_time: typing.Optional[analytics_spec.Timestamp] = None,
) -> analytics_spec.Timestamps:

    if end_time is None:
        end_time = int(time.time())

    return tooltime.get_standard_intervals(
        interval_size=timescale['interval_size'],
        window_size=timescale['window_size'],
        end_time=end_time,
    )


async def async_get_timestamps_blocks(
    timestamps: typing.Sequence[analytics_spec.Timestamp],
    provider: spec.ProviderSpec,
    **kwargs
) -> list[int]:

    # use latest block if last timestamp is greater than latest block timestamp
    latest_block = await evm.async_get_block('latest', provider=provider)
    if latest_block['timestamp'] <= timestamps[-1]:
        timestamps = timestamps[:-1]
        latest_block_swapped = True
    else:
        latest_block_swapped = False

    # allow only the last timestamp to exceed the latest block time
    for timestamp in timestamps:
        if timestamp > latest_block['timestamp']:
            raise Exception('timestamps are greater that latest block time')

    # get blocks of timestamps
    blocks = await evm.async_get_blocks_of_timestamps(
        timestamps, provider=provider, **kwargs
    )

    # append latest block if swapped in for last timestamp
    if latest_block_swapped:
        blocks.append(latest_block['number'])

    return blocks


def summarize_timestamps(
    timescale: analytics_spec.Timescale,
    timestamps: analytics_spec.Timestamps,
) -> None:
    toolstr.print_text_box(
        'window = '
        + tooltime.timelength_to_phrase(timescale['window_size'])
        + ', precision = '
        + tooltime.timelength_to_phrase(timescale['interval_size'])
        + ', n_points = '
        + str(len(timestamps))
    )
    n_show = 2
    if len(timestamps) > n_show * 2:
        for timestamp in timestamps[:n_show]:
            print('   ', tooltime.timestamp_to_iso(timestamp))
        print('   ', '...')
        for timestamp in timestamps[-n_show:]:
            print('   ', tooltime.timestamp_to_iso(timestamp))
    else:
        for timestamp in timestamps:
            print('   ', tooltime.timestamp_to_iso(timestamp))
    print()
    print()




from ctc import binary
from ctc import spec
from ctc import rpc
from ctc.toolbox import backend_utils
from ctc.toolbox import pd_utils

from .event_backends import filesystem_events
from .event_backends import node_events
from .. import block_utils
from .. import abi_utils


def get_backend_functions():
    return {
        'get': {
            'filesystem': filesystem_events.async_get_events_from_filesystem,
            'download': async_download_events,
            'node': node_events.async_get_events_from_node,
        },
        'save': {
            'filesystem': filesystem_events.async_save_events_to_filesystem,
        },
    }


async def async_get_events(
    *,
    contract_address,
    start_block=None,
    end_block=None,
    backend_order=None,
    keep_multiindex=True,
    **query
) -> spec.DataFrame:

    if start_block is None:
        start_block = await block_utils.async_get_contract_creation_block(
            contract_address
        )
    if end_block is None:
        end_block = 'latest'

    start_block, end_block = await block_utils.async_block_numbers_to_int(
        blocks=[start_block, end_block],
    )

    if backend_order is None:
        backend_order = ['filesystem', 'download']

    events = await backend_utils.async_run_on_backend(
        get_backend_functions()['get'],
        contract_address=contract_address,
        start_block=start_block,
        end_block=end_block,
        backend_order=backend_order,
        **query
    )

    if not keep_multiindex:
        events.index = pd_utils.keep_level(
            index=events.index, level='block_number'
        )

    return events


async def async_save_events(events, **query):
    return await backend_utils.async_run_on_backend(
        get_backend_functions()['save'], events=events, **query
    )


async def async_transfer_events(
    *, contract_address, start_block=None, end_block=None, **query
):

    start_block, end_block = await block_utils.async_block_numbers_to_int(
        blocks=[start_block, end_block],
    )

    return await backend_utils.async_transfer_backends(
        get=async_get_events,
        save=async_save_events,
        contract_address=contract_address,
        start_block=start_block,
        end_block=end_block,
        **query
    )


async def async_download_events(
    contract_address,
    event_hash=None,
    event_name=None,
    event_abi=None,
    start_block=None,
    end_block=None,
    provider=None,
    verbose=True,
):

    if event_hash is None and event_name is None and event_abi is None:
        raise Exception('must specify either event_hash or event_name')

    contract_address = contract_address.lower()

    start_block, end_block = await block_utils.async_block_numbers_to_int(
        blocks=[start_block, end_block],
        provider=provider,
    )

    provider = rpc.get_provider(provider)
    network = provider['network']
    if network is None:
        raise Exception('could not determine network')

    # get event hash
    if event_hash is None:
        if event_abi is None:
            if event_name is None:
                raise Exception('must specify more event information')
            event_abi = await abi_utils.async_get_event_abi(
                contract_address=contract_address,
                event_name=event_name,
                network=network,
            )

        event_hash = binary.get_event_hash(event_abi)

    # determine what needs to be downloaded
    listed_events = filesystem_events.list_events(
        contract_address=contract_address,
        event_hash=event_hash,
    )
    downloads = []
    if listed_events is None:
        download = {'start_block': start_block, 'end_block': end_block}
        downloads.append(download)
    else:

        block_range = listed_events['block_range']
        if start_block < block_range[0]:
            download = {
                'start_block': start_block,
                'end_block': block_range[0] - 1,
                'common_kwargs': {'verbose': verbose},
            }
            downloads.append(download)
        if end_block > block_range[-1]:
            download = {
                'start_block': block_range[-1] + 1,
                'end_block': end_block,
                'common_kwargs': {'verbose': verbose},
            }
            downloads.append(download)

    # perform downloads
    for download in downloads:
        await async_transfer_events(
            from_backend='node',
            to_backend='filesystem',
            event_hash=event_hash,
            event_abi=event_abi,
            contract_address=contract_address,
            provider=provider,
            **download
        )

    # load from filesystem
    return await filesystem_events.async_get_events_from_filesystem(
        event_hash=event_hash,
        event_abi=event_abi,
        contract_address=contract_address,
        start_block=start_block,
        end_block=end_block,
        verbose=verbose,
        provider=provider,
    )




from __future__ import annotations

import typing

from ctc import spec

from .. import chainlink_metadata
from .. import chainlink_spec
from . import feed_datum_by_block
from . import feed_events


@typing.overload
async def async_get_feed_data(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['full'],
    blocks: typing.Sequence[spec.BlockNumberReference] = None,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = True,
    interpolate: bool = False,
    provider: spec.ProviderSpec = None,
) -> spec.DataFrame:
    ...


@typing.overload
async def async_get_feed_data(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['answer'] = 'answer',
    blocks: typing.Sequence[spec.BlockNumberReference] = None,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = True,
    interpolate: bool = False,
    provider: spec.ProviderSpec = None,
) -> spec.Series:
    ...


async def async_get_feed_data(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['answer', 'full'] = 'answer',
    blocks: typing.Sequence[spec.BlockNumberReference] = None,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = True,
    interpolate: bool = False,
    provider: spec.ProviderSpec = None,
) -> typing.Union[spec.DataFrame, spec.Series]:

    # determine blocks
    if (blocks is not None) and [start_block, end_block].count(None) < 2:
        raise Exception('should only provide one block specification')
    if blocks is None:
        if start_block is None:
            start_block = await chainlink_metadata.async_get_feed_first_block(
                feed=feed,
                provider=provider,
            )
        if end_block is None:
            end_block = 'latest'

    # perform query
    if fields == 'answer':

        if blocks is not None:
            return (
                await feed_datum_by_block.async_get_feed_answer_datum_by_block(
                    feed,
                    provider=provider,
                    blocks=blocks,
                    normalize=normalize,
                    interpolate=interpolate,
                )
            )

        else:
            return await feed_events.async_get_answer_feed_event_data(
                feed,
                provider=provider,
                normalize=normalize,
                start_block=start_block,
                end_block=end_block,
                interpolate=interpolate,
            )

    elif fields == 'full':

        if blocks is not None:
            return await feed_datum_by_block.async_get_feed_full_datum_by_block(
                feed,
                provider=provider,
                blocks=blocks,
                normalize=normalize,
                interpolate=interpolate,
            )

        else:
            return await feed_events.async_get_full_feed_event_data(
                feed,
                provider=provider,
                normalize=normalize,
                start_block=start_block,
                end_block=end_block,
                interpolate=interpolate,
            )

    else:
        raise Exception('unknown fields format: ' + str(fields))




import pytest
import toolconfig

from ctc import spec
from ctc import directory


@pytest.mark.parametrize(
    'test',
    [
        ['chainlink', 'ETH_USD', '0x5f4ec3df9cbd43714fe2740f5e3616155c5b8419'],
    ],
)
def test_get_oracle_address(test):
    protocol, feed_name, feed_address = test
    actual_address = directory.get_oracle_address(
        name=feed_name,
        protocol=protocol,
    )
    assert actual_address == feed_address


def test_get_oracle_feed_metadata():
    metadata = directory.get_oracle_feed_metadata(name='ETH_USD')
    toolconfig.conforms_to_spec(metadata, spec.OracleFeedMetadata)


def test_load_oracle_feeds():
    oracle_feeds = directory.load_oracle_feeds()
    for metadata in oracle_feeds.values():
        toolconfig.conforms_to_spec(metadata, spec.OracleFeedMetadata)


def test_get_oracle_feed_path():
    path = directory.get_oracle_feed_path()
    assert isinstance(path, str) and len(path) > 0




import os
import typing
from typing_extensions import Literal

import toolcache


formats = ['json', 'ast', 'csv', 'yaml', 'toml']
DataFormat = Literal['json', 'ast', 'csv', 'yaml', 'toml']


def get_path_data_format(path: str) -> DataFormat:

    extension = os.path.splitext(path)[-1]
    extension = extension[1:]

    # python3.7 compatibility
    # args = typing.get_args(DataFormat)
    args = DataFormat.__args__  # type: ignore

    if extension in args:
        return typing.cast(DataFormat, extension)
    else:
        raise Exception('unknown file type: ' + str(extension))


@toolcache.cache(cachetype='memory')
def load_file_data(
    path: str, format: typing.Optional[DataFormat] = None
) -> typing.Any:

    if format is None:
        format = get_path_data_format(path=path)
    with open(path, 'r') as f:
        return load_buffer_data(buffer=f, format=format)


def load_buffer_data(buffer: typing.TextIO, format: DataFormat) -> typing.Any:
    if format == 'json':
        import json

        return json.load(buffer)
    elif format == 'ast':
        import ast

        return ast.literal_eval(buffer.read())
    elif format == 'csv':
        import pandas as pd

        df = pd.read_csv(buffer)
        return df.to_dict(orient='records')
    elif format == 'yaml':
        import yaml

        return yaml.safe_load(buffer)
    elif format == 'toml':
        import toml

        return toml.load(buffer)
    else:
        raise Exception('unknown format: ' + str(format))


def write_file_data(
    path: str,
    data: typing.Union[
        typing.Sequence[typing.Mapping],
        typing.Mapping[typing.Any, typing.Mapping],
    ],
    format: typing.Optional[DataFormat] = None,
    overwrite: bool = False,
    index_field: typing.Optional[str] = None,
) -> None:

    if format is None:
        format = get_path_data_format(path=path)

    if os.path.isfile(path) and not overwrite:
        raise Exception('file already exists, use overwrite=True')

    with open(path, 'w') as f:
        write_buffer_data(
            buffer=f,
            data=data,
            index_field=index_field,
            format=format,
        )


def write_buffer_data(
    buffer: typing.TextIO,
    data: typing.Union[
        typing.Sequence[typing.Mapping],
        typing.Mapping[typing.Any, typing.Mapping],
    ],
    format: DataFormat,
    index_field: typing.Optional[str] = None,
) -> None:

    if index_field is not None:
        if not isinstance(data, dict):
            raise Exception('index_field is used to convert dict -> list')
        data = [
            dict(datum, **{index_field: index}) for index, datum in data.items()
        ]

    if format == 'json':
        import json

        json.dump(data, buffer)
    elif format == 'ast':
        buffer.write(str(data))
    elif format == 'toml':
        import toml

        if not isinstance(data, typing.Mapping):
            raise Exception('can only write mappings to toml, not sequences')

        toml.dump(data, buffer)
    elif format == 'yaml':
        import yaml

        yaml.dump(data, buffer)
    elif format == 'csv':
        import csv

        # gather fields
        field_set: set[str] = set()
        if isinstance(data, list):
            dataiter: typing.Iterable[dict] = data
        elif isinstance(data, dict):
            dataiter = data.values()
        else:
            raise Exception('unknown data type: ' + str(type(data)))
        field_set = {key for datum in dataiter for key in datum.keys()}
        fields = sorted(field_set)

        writer = csv.DictWriter(buffer, fieldnames=fields)
        writer.writeheader()
        writer.writerows(data)
    else:
        raise Exception('unknown format: ' + str(format))




"""
# balances of wallets within single block
ctc eth balances <wallets> [--block <block>]

# balance of single wallet across multiple blocks
ctc eth balances <wallet> [--blocks <block>]
"""

import pandas as pd

from ctc import evm
from ctc import rpc
from ctc.cli import cli_utils


def get_command_spec():
    return {
        'f': async_balances_command,
        'args': [
            {'name': 'wallets', 'kwargs': {'nargs': '+'}},
            {'name': '--block'},
            {'name': '--blocks', 'kwargs': {'nargs': '+'}},
            #
            {'name': '--raw', 'kwargs': {'action': 'store_true'}},
            {'name': '--output', 'kwargs': {'default': 'stdout'}},
            {'name': '--overwrite', 'kwargs': {'action': 'store_true'}},
        ],
    }


async def async_balances_command(
    wallets, block, blocks, raw, output, overwrite,
):
    indent = None
    wallets = [wallet.lower() for wallet in wallets]

    if blocks is not None:
        # single wallet, multiple blocks
        if len(wallets) > 1:
            raise Exception('cannot specify multiple wallets and multiple blocks')

        wallet = wallets[0]
        resolved_blocks = await cli_utils.async_resolve_blocks(blocks)
        balances = await evm.async_get_eth_balance_by_block(
            address=wallet,
            blocks=resolved_blocks,
            normalize=(not raw),
        )
        df = pd.DataFrame(balances, index=resolved_blocks)
        df.index.name = 'block'
        df.columns = ['balance']

    else:
        # multiple wallets, single block
        if blocks is not None:
            raise Exception('cannot specify multiple wallets and multiple blocks')

        if block is None:
            block = 'latest'
        block = await evm.async_block_number_to_int(block)

        if wallets is not None:
            balances = await evm.async_get_eth_balance_of_addresses(
                addresses=wallets,
                block=block,
                normalize=(not raw),
            )

            df = pd.Series(balances, index=wallets)
            df.name = 'balance'
            df.index.name = 'address'

    cli_utils.output_data(df, output, overwrite, indent=indent, raw=raw)

    await rpc.async_close_http_session()




import pytest
import toolconfig

from ctc import spec
from ctc import directory


@pytest.mark.parametrize(
    'query',
    [
        {'symbol': 'USDC'},
        {'address': '0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48'},
        {'address': '0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48'},
    ],
)
def test_get_erc20_metadata(query):
    metadata = directory.get_erc20_metadata(network='mainnet', **query)
    toolconfig.conforms_to_spec(metadata, spec.ERC20Metadata)


@pytest.mark.parametrize(
    'test',
    [
        ['USDC', '0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48'],
    ],
)
def test_get_erc20_address(test):
    symbol, address = test
    actual_address = directory.get_erc20_address(symbol=symbol)
    assert address == actual_address


@pytest.mark.parametrize(
    'test',
    [['0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48', 'USDC']],
)
def test_get_erc20_symbol(test):
    address, symbol = test
    actual_symbol = directory.get_erc20_symbol(address=address)
    assert symbol == actual_symbol


def test_load_filesystem_erc20s_data():
    erc20s_data = directory.load_filesystem_erc20_data('mainnet')
    for erc20_data in erc20s_data:
        toolconfig.conforms_to_spec(erc20_data, spec.ERC20Metadata)




import filecmp
import os
import shutil

import toolcli

import ctc


def get_default_data_dir() -> str:
    return os.path.abspath(os.path.join(ctc.__path__[0], 'default_data'))


def is_data_root_initialized(data_root_path: str) -> bool:
    """a data root is considered initialized if it contains all default data"""

    data_root_path = os.path.abspath(data_root_path)
    default_data_dir = get_default_data_dir()

    for root, subdirs, files in os.walk(default_data_dir):

        if root == default_data_dir:
            check_root = data_root_path
        else:
            root_relpath = os.path.relpath(root, default_data_dir)
            check_root = os.path.join(data_root_path, root_relpath)

        for subdir in subdirs:
            subdir_path = os.path.join(check_root, subdir)
            if not os.path.isdir(subdir_path):
                return False
        for file in files:
            default_data_file_path = os.path.join(root, file)
            file_path = os.path.join(check_root, file)
            if not os.path.isfile(file_path) or not filecmp.cmp(
                file_path, default_data_file_path
            ):
                return False

    return True


def initialize_data_root(
    path: str, confirm: bool = False, raise_if_unconfirmed: bool = True
) -> bool:

    default_data_dir = get_default_data_dir()

    # validate directory name
    if os.path.splitext(path)[-1] != '':
        raise Exception('must use a directory path, not a file path')

    print()
    print('Will use data root:', path)
    if not os.path.isdir(path):
        if not confirm:
            print()
            answer = toolcli.input_yes_or_no(
                'Directory does not exist. Create it?', default='yes'
            )
            if not answer:
                if raise_if_unconfirmed:
                    raise Exception('must create directory')
                else:
                    return False

    else:
        overwritten = []
        new_files = []
        unchanged_files = []
        for root, subdirs, files in os.walk(default_data_dir):

            if root == default_data_dir:
                check_root = path
            else:
                root_relpath = os.path.relpath(root, default_data_dir)
                check_root = os.path.join(path, root_relpath)

            # root_relpath = os.path.relpath(root, default_data_dir)
            # check_root = os.path.join(path, root_relpath)
            for file in files:
                filepath = os.path.join(check_root, file)
                if os.path.isfile(filepath):
                    with open(filepath, 'r') as f:
                        old_file_data = f.read()
                    with open(os.path.join(root, file), 'r') as f:
                        new_file_data = f.read()
                    if old_file_data == new_file_data:
                        unchanged_files.append(filepath)
                    else:
                        overwritten.append(filepath)
                else:
                    new_files.append(filepath)

        if len(new_files) > 0:
            print()
            print('Writing', len(new_files), 'new files')
            for new_file in new_files:
                print('-', new_file)

        if len(overwritten) > 0:
            print()
            print('Will overwrite the following files:')
            for filepath in overwritten:
                print('-', filepath)
            print()
            answer = toolcli.input_yes_or_no(
                'Continue? ', default='yes', default_prefix='(default = '
            )
            if not answer:
                if raise_if_unconfirmed:
                    raise Exception('Must overwrite files to continue')
                else:
                    return False

    # create directory
    shutil.copytree(default_data_dir, path, dirs_exist_ok=True)
    return True




import asyncio

from ctc.protocols import chainlink_utils

from . import pool_metadata
from . import token_state


async def async_get_pool_prices(
    *,
    oracle=None,
    ctokens=None,
    comptroller=None,
    block='latest',
    to_usd=True,
):
    if oracle is None:
        if comptroller is None:
            raise Exception('specify comptroller')
        oracle = await pool_metadata.async_get_pool_oracle(
            comptroller, block=block
        )
    if ctokens is None:
        if comptroller is None:
            raise Exception('specify comptroller')
        ctokens = await pool_metadata.async_get_pool_ctokens(
            comptroller, block=block
        )

    coroutines = [
        token_state.async_get_ctoken_price(
            ctoken=ctoken, oracle=oracle, block=block
        )
        for ctoken in ctokens
    ]
    prices = await asyncio.gather(*coroutines)

    if to_usd:
        eth_price = await chainlink_utils.async_get_eth_price(block=block)
        prices = [price * eth_price for price in prices]

    return dict(zip(ctokens, prices))


async def get_pool_tvl_and_tvb(
    *, comptroller=None, ctokens=None, oracle=None, block='latest'
):
    if ctokens is None:
        if comptroller is None:
            raise Exception(
                'must specify comptroller if not specifying ctokens'
            )
        ctokens = await pool_metadata.async_get_pool_ctokens(
            comptroller=comptroller, block=block
        )
    if oracle is None:
        if comptroller is None:
            raise Exception('must specify comptroller if not specifying oracle')
        oracle = await pool_metadata.async_get_pool_oracle(
            comptroller=comptroller, block=block
        )

    eth_price = await chainlink_utils.async_get_eth_price(block=block)

    ctokens_stats = [
        asyncio.create_task(
            token_state.async_get_ctoken_tvl_and_tvb(
                ctoken, oracle, eth_price, block=block
            )
        )
        for ctoken in ctokens
    ]
    ctokens_stats = await asyncio.gather(*ctokens_stats)

    tvl = 0
    tvb = 0
    for ctoken_stats in ctokens_stats:
        tvl += ctoken_stats['tvl']
        tvb += ctoken_stats['tvb']

    return {'tvl': tvl, 'tvb': tvb}




from __future__ import annotations

from ctc import rpc
from ctc import spec
from . import block_normalize


async def async_get_block(
    block: spec.BlockReference,
    include_full_transactions: bool = False,
    provider: spec.ProviderSpec = None,
) -> spec.Block:

    if spec.is_block_number_reference(block):

        return await rpc.async_eth_get_block_by_number(
            block_number=block_normalize.standardize_block_number(block),
            provider=provider,
            include_full_transactions=include_full_transactions,
        )

    elif spec.is_block_hash(block):

        return await rpc.async_eth_get_block_by_hash(
            block_hash=block,
            provider=provider,
            include_full_transactions=include_full_transactions,
        )

    else:
        raise Exception('unknown block specifier: ' + str(block))


async def async_get_blocks(
    blocks: list[spec.BlockReference],
    include_full_transactions: bool = False,
    chunk_size: int = 500,
    provider: spec.ProviderSpec = None,
) -> list[spec.Block]:

    provider = rpc.add_provider_parameters(provider, {'chunk_size': chunk_size})

    if all(spec.is_block_number_reference(block) for block in blocks):

        standardized = [
            block_normalize.standardize_block_number(block) for block in blocks
        ]

        return await rpc.async_batch_eth_get_block_by_number(
            block_numbers=standardized,
            include_full_transactions=include_full_transactions,
            provider=provider,
        )

    elif all(spec.is_block_hash(block) for block in blocks):

        return await rpc.async_batch_eth_get_block_by_number(
            block_hashes=blocks,
            include_full_transactions=include_full_transactions,
            provider=provider,
        )

    else:
        raise Exception(
            'blocks should be all block number references or all block hashes'
        )


async def async_get_latest_block_number(
    provider: spec.ProviderSpec = None,
) -> int:
    return await rpc.async_eth_block_number(provider=provider)


async def async_get_blocks_timestamps(
    blocks: list[spec.BlockReference], **get_blocks_kwargs
) -> list[int]:
    return [
        block['timestamp']
        for block in await async_get_blocks(blocks=blocks, **get_blocks_kwargs)
    ]




from __future__ import annotations

import typing

from ctc import directory
from ctc import spec
from ctc.toolbox import async_utils
from ctc.toolbox import nested_utils

from . import coracle_spec
from . import coracle_deposits
from . import coracle_balances


#
# # fei deposits
#


async def async_get_fei_deposit_balances(
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
) -> dict[spec.ContractAddress, typing.Union[int, float]]:

    FEI = directory.get_erc20_address(symbol='FEI')
    fei_deposits = await coracle_deposits.async_get_token_deposits(
        token=FEI,
        block=block,
        provider=provider,
    )

    fei_balances = await coracle_balances.async_get_deposits_balances(
        deposits=fei_deposits,
        block=block,
        provider=provider,
    )

    non_fei_balances = await _async_get_non_fei_deposits_fei_balances(
        block=block, provider=provider
    )
    fei_deposits = fei_deposits + tuple(non_fei_balances.keys())
    fei_balances.extend(non_fei_balances.values())

    result: typing.Union[list[int], list[float]]
    if normalize:
        result = [balance / 1e18 for balance in fei_balances]
    else:
        result = fei_balances

    return dict(zip(fei_deposits, result))


async def _async_get_non_fei_deposits_fei_balances(block, provider=None):

    tokens_deposits = await coracle_deposits.async_get_tokens_deposits(
        block=block
    )

    non_fei_deposits = []
    for token, deposits in tokens_deposits.items():
        non_fei_deposits.extend(deposits)

    balances = (
        await coracle_balances.async_get_deposits_resistant_balances_and_fei(
            deposits=non_fei_deposits,
            block=block,
        )
    )

    fei_balances = [balance[1] for balance in balances]

    return dict(zip(non_fei_deposits, fei_balances))


async def async_get_fei_deposit_balances_by_block(
    blocks: typing.Sequence[spec.BlockNumberReference],
    provider: spec.ProviderSpec = None,
) -> dict[spec.ContractAddress, list[typing.Union[int, float]]]:

    coroutines = [
        async_get_fei_deposit_balances(block=block, provider=provider)
        for block in blocks
    ]
    results = await async_utils.gather_coroutines(*coroutines)
    return nested_utils.list_of_dicts_to_dict_of_lists(results)


#
# # fei platforms
#

T = typing.TypeVar('T', int, float)


def fei_deposits_to_deployments(deposit_balances: dict[str, T]) -> dict[str, T]:

    deployment_balances: dict[str, T] = {}
    for deposit, value in deposit_balances.items():

        if deposit in coracle_spec.deposit_metadata:
            deployment = coracle_spec.deposit_metadata[deposit]['platform']
        else:
            deployment = 'Other'

        deployment_balances.setdefault(deployment, type(value)())
        deployment_balances[deployment] += value

    return deployment_balances


def fei_deposits_to_deployments_by_block(
    deposit_balances_by_block: dict[str, list[T]]
) -> dict[str, list[T]]:

    deployment_balances_by_block: dict[str, list[T]] = {}

    for deposit, value in deposit_balances_by_block.items():

        if deposit in coracle_spec.deposit_metadata:
            deployment = coracle_spec.deposit_metadata[deposit]['platform']
        else:
            deployment = 'Other'

        n_blocks = len(value)
        empty = [type(value[i])() for i in range(n_blocks)]
        deployment_balances_by_block.setdefault(deployment, empty)
        deployment_balances_by_block[deployment] = [
            lhs + rhs
            for lhs, rhs in zip(value, deployment_balances_by_block[deployment])
        ]

    return deployment_balances_by_block




import typing

from ctc import directory
from ctc import evm
from ctc import rpc
from ctc import spec

from . import chainlink_spec


def get_feed_decimals(
    feed: chainlink_spec._FeedReference,
    provider: spec.ProviderSpec = None,
):
    provider = rpc.get_provider(provider)
    network = provider['network']
    if network is None:
        raise Exception('could not determine network')
    metadata = directory.get_oracle_feed_metadata(
        address=feed,
        network=network,
        protocol='chainlink',
    )
    return metadata['decimals']


async def async_resolve_feed_address(
    feed: str,
    *,
    provider: spec.ProviderSpec = None,
    block: spec.BlockNumberReference = 'latest',
) -> spec.Address:

    provider = rpc.get_provider(provider)
    network = provider['network']
    if network is None:
        raise Exception('could not determine network')

    if evm.is_address_str(feed):
        return feed
    elif isinstance(feed, str):
        return directory.get_oracle_address(
            name=feed,
            protocol='chainlink',
            network=network,
            block=block,
        )
    else:
        raise Exception('invalid feed reference: ' + str(feed))


async def async_get_feed_aggregator(
    feed: chainlink_spec._FeedReference,
    *,
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
) -> spec.Address:

    feed = await async_resolve_feed_address(
        feed, block=block, provider=provider
    )

    return await rpc.async_eth_call(
        to_address=feed,
        function_name='aggregator',
        block_number=block,
    )


async def async_get_feed_first_block(
    feed: chainlink_spec._FeedReference,
    provider: spec.ProviderSpec = None,
    start_search: typing.Optional[spec.BlockNumberReference] = None,
    end_search: typing.Optional[spec.BlockNumberReference] = None,
    verbose: bool = False,
) -> int:

    feed = await async_resolve_feed_address(feed, provider=provider)

    return await evm.async_get_contract_creation_block(
        contract_address=feed,
        start_block=start_search,
        end_block=end_search,
        verbose=verbose,
    )




from __future__ import annotations

import typing

from ctc import rpc
from ctc import spec

from .. import chainlink_metadata
from .. import chainlink_spec


@typing.overload
async def async_get_feed_datum(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['full'],
    normalize: typing.Literal[False],
    block: typing.Optional[spec.BlockNumberReference] = None,
    provider: spec.ProviderSpec = None,
) -> chainlink_spec.FeedRoundData:
    ...


@typing.overload
async def async_get_feed_datum(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['full'],
    normalize: typing.Literal[True],
    block: typing.Optional[spec.BlockNumberReference] = None,
    provider: spec.ProviderSpec = None,
) -> chainlink_spec.FeedRoundDataNormalized:
    ...


# overloading in python is a mess...
@typing.overload
async def async_get_feed_datum(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['full'],
    normalize: bool = True,
    block: typing.Optional[spec.BlockNumberReference] = None,
    provider: spec.ProviderSpec = None,
) -> typing.Union[
    chainlink_spec.FeedRoundData, chainlink_spec.FeedRoundDataNormalized
]:
    ...


@typing.overload
async def async_get_feed_datum(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['answer'] = 'answer',
    normalize: typing.Literal[False],
    block: typing.Optional[spec.BlockNumberReference] = None,
    provider: spec.ProviderSpec = None,
) -> int:
    ...


@typing.overload
async def async_get_feed_datum(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['answer'] = 'answer',
    normalize: bool = True,
    block: typing.Optional[spec.BlockNumberReference] = None,
    provider: spec.ProviderSpec = None,
) -> float:
    ...


async def async_get_feed_datum(
    feed: chainlink_spec._FeedReference,
    *,
    fields: typing.Literal['answer', 'full'] = 'answer',
    normalize: bool = True,
    block: typing.Optional[spec.BlockNumberReference] = None,
    provider: spec.ProviderSpec = None,
) -> typing.Union[
    int,
    float,
    chainlink_spec.FeedRoundData,
    chainlink_spec.FeedRoundDataNormalized,
]:
    """get feed data for a single block"""

    if block is None:
        block = 'latest'

    feed = await chainlink_metadata.async_resolve_feed_address(feed)

    if fields == 'answer':

        answer = await rpc.async_eth_call(
            to_address=feed,
            function_name='latestAnswer',
            block_number=block,
            provider=provider,
            fill_empty=True,
            empty_token=None,
        )

        if normalize and answer is not None:
            decimals = chainlink_metadata.get_feed_decimals(feed)
            answer /= 10 ** decimals

        return answer

    elif fields == 'full':

        data = await rpc.async_eth_call(
            to_address=feed,
            function_name='latestRoundData',
            block_number=block,
            provider=provider,
            fill_empty=True,
            empty_token=None,
        )

        round_id, answer, _started_at, updated_at, _answered_in_round_id = data

        full: chainlink_spec.FeedRoundData = {
            'answer': answer,
            'timestamp': updated_at,
            'round_id': round_id,
        }

        if normalize and answer is not None:
            decimals = chainlink_metadata.get_feed_decimals(feed)
            full['answer'] /= 10 ** decimals
            return full

        else:

            return full

    else:
        raise Exception('unknown fields type: ' + str(fields))




from __future__ import annotations

import typing

from ctc import evm
from ctc import spec


async def async_resolve_block_range(
    block_range: typing.Union[str, typing.Sequence[str], spec.BlockRange],
) -> tuple[int, int]:

    if not isinstance(block_range, dict):
        block_range = parse_block_range(block_range)

    start_block, end_block = await evm.async_block_numbers_to_int(
        [block_range['start_block'], block_range['end_block']],
    )

    if block_range['open_start']:
        start_block = start_block + 1
    if block_range['open_end']:
        end_block = end_block - 1

    if start_block > end_block:
        raise Exception('invalid block range: ' + str([start_block, end_block]))

    return start_block, end_block


async def async_resolve_block_sample(
    block_sample: typing.Union[str, typing.Sequence[str], spec.BlockSample],
) -> list[int]:

    if not isinstance(block_sample, dict):
        block_sample = parse_block_sample(block_sample)

    start_block, end_block = await evm.async_block_numbers_to_int(
        [block_sample['start_block'], block_sample['end_block']],
    )

    if block_sample['open_start']:
        start_block = start_block + 1
    if block_sample['open_end']:
        end_block = end_block - 1

    if start_block > end_block:
        raise Exception('invalid block range: ' + str([start_block, end_block]))

    block_interval = block_sample['block_interval']
    if block_interval is None:
        block_interval = 1
    return list(range(start_block, end_block + 1, block_interval))


def is_block_range(block_spec: typing.Union[str, typing.Sequence[str]]) -> bool:
    try:
        parse_block_range(block_spec)
        return True
    except ValueError:
        return False


def is_block_sample(
    block_spec: typing.Union[str, typing.Sequence[str]]
) -> bool:
    try:
        parse_block_sample(block_spec)
        return True
    except ValueError:
        return False


def parse_block_range(
    block_spec: typing.Union[str, typing.Sequence[str]]
) -> spec.BlockRange:
    """
    ## Example: All of the following should produce the same output
    - "[123 456]"
    - "[123,456]"
    - "[123, 456]"
    - "[123 , 456]"
    - "[123 ,456]"
    - "[123  456]"
    - "[123,  456]"
    - "[123  ,  456]"
    - "[123  ,456]"
    - "[ 123 456]"
    - "[ 123,456   ]"
    - "[ 123, 456]"
    - "[ 123 , 456]"
    - "[ 123 ,456]"
    """

    tokens = tokenize(block_spec)

    # parse out the four components
    if len(tokens) != 4:
        raise ValueError(
            'invalid block range specification: ' + str(block_spec)
        )
    start_type, start_block, end_block, end_type = tokens

    # convert to int if applicable
    if start_block not in spec.block_number_names:
        start_block = evm.raw_block_number_to_int(start_block)
    if end_block not in spec.block_number_names:
        end_block = evm.raw_block_number_to_int(end_block)

    # detect interval types
    if start_type == '(':
        open_start = True
    elif start_type == '[':
        open_start = False
    else:
        raise ValueError(
            'invalid interval type in block range: ' + str(block_spec)
        )
    if start_type == '(':
        open_end = True
    elif start_type == '[':
        open_end = False
    else:
        raise ValueError(
            'invalid interval type in block range: ' + str(block_spec)
        )

    return {
        'start_block': start_block,
        'end_block': end_block,
        'open_start': open_start,
        'open_end': open_end,
    }


def parse_block_sample(block_spec):

    tokens = tokenize(block_spec)

    # parse out the four components
    if len(tokens) == 4:
        start_type, start_block, end_block, end_type = tokens
        block_interval = None
    elif len(tokens) == 5:
        start_type, start_block, end_block, block_interval, end_type = tokens
    else:
        raise ValueError(
            'invalid block range specification: ' + str(block_spec)
        )

    # convert to int if applicable
    if start_block not in spec.block_number_names:
        start_block = evm.raw_block_number_to_int(start_block)
    if end_block not in spec.block_number_names:
        end_block = evm.raw_block_number_to_int(end_block)
    if block_interval is not None:
        block_interval = evm.raw_block_number_to_int(block_interval)

    # detect interval types
    if start_type == '(':
        open_start = True
    elif start_type == '[':
        open_start = False
    else:
        raise ValueError(
            'invalid interval type in block range: ' + str(block_spec)
        )
    if end_type == ')':
        open_end = True
    elif end_type == ']':
        open_end = False
    else:
        raise ValueError(
            'invalid interval type in block range: ' + str(block_spec)
        )

    return {
        'start_block': start_block,
        'end_block': end_block,
        'block_interval': block_interval,
        'open_start': open_start,
        'open_end': open_end,
    }


def tokenize(data: typing.Union[str, typing.Sequence[str]]) -> list[str]:
    """break up string into tokens, tokens can be separated by commas or spaces

    creates separate tokens for:
    - "(" or "[" at beginning
    - ")" or "]" at end
    """

    # break into tokens
    if isinstance(data, str):
        data = [data]
    tokens = []
    for datum in data:
        datum = datum.replace(',', ' ')
        subtokens = datum.split(' ')

        for token in subtokens:
            if len(token) == 0:
                continue
            elif len(token) == 1:
                tokens.append(token)
            else:

                start_interval = token[0] in ['(', '[']
                end_interval = token[-1] in [')', ']']

                # convert token based on contained intervals
                if start_interval and end_interval:
                    tokens.append(token[0])
                    if len(token) > 2:
                        tokens.append(token[1:-1])
                    tokens.append(token[-1])
                elif start_interval:
                    tokens.append(token[0])
                    tokens.append(token[1:])
                elif end_interval:
                    tokens.append(token[:-1])
                    tokens.append(token[-1])
                else:
                    tokens.append(token)

    return tokens




import toolcli

command_index = {
    #
    # admin commands
    (): 'ctc.cli.commands.root_command',
    ('cd',): 'ctc.cli.commands.admin.cd_command',
    ('config',): 'ctc.cli.commands.admin.config_command',
    ('config', 'edit'): 'ctc.cli.commands.admin.config.edit_command',
    ('config', 'path'): 'ctc.cli.commands.admin.config.path_command',
    ('download-proxy-abi',): 'ctc.cli.commands.admin.download_proxy_abi_command',
    ('setup',): 'ctc.cli.commands.admin.setup_command',
    ('rechunk-events',): 'ctc.cli.commands.admin.rechunk_command',
    #
    # compute commands
    ('keccak',): 'ctc.cli.commands.compute.keccak_command',
    #
    # data commands
    ('address',): 'ctc.cli.commands.data.address_command',
    ('block',): 'ctc.cli.commands.data.block_command',
    ('blocks',): 'ctc.cli.commands.data.blocks_command',
    ('call',): 'ctc.cli.commands.data.call_command',
    ('calls',): 'ctc.cli.commands.data.calls_command',
    ('eth', 'balance'): 'ctc.cli.commands.data.eth.balance_command',
    ('eth', 'balances'): 'ctc.cli.commands.data.eth.balances_command',
    ('erc20', 'balance'): 'ctc.cli.commands.data.erc20.balance_command',
    ('erc20', 'balances'): 'ctc.cli.commands.data.erc20.balances_command',
    ('erc20', 'transfers'): 'ctc.cli.commands.data.erc20.transfers_command',
    ('events',): 'ctc.cli.commands.data.events_command',
    ('gas',): 'ctc.cli.commands.data.gas_command',
    ('transaction',): 'ctc.cli.commands.data.transaction_command',
    #
    # protocol commands
    ('chainlink',): 'ctc.protocols.chainlink_utils.cli.chainlink_command',
    ('fei', 'payload'): 'ctc.protocols.fei_utils.cli.fei.payload_command',
    ('rari', 'fuse'): 'ctc.protocols.rari_utils.cli.rari.fuse_command',
    ('uniswap', 'mints'): 'ctc.protocols.uniswap_v2_utils.cli.mints_command',
    ('uniswap', 'burns'): 'ctc.protocols.uniswap_v2_utils.cli.burns_command',
    ('uniswap', 'swaps'): 'ctc.protocols.uniswap_v2_utils.cli.swaps_command',
    ('uniswap', 'pool'): 'ctc.protocols.uniswap_v2_utils.cli.pool_command',
}


def run_cli(raw_command=None, **toolcli_kwargs):

    config = {
        'include_cd': True,
        'include_debug': True,
    }
    toolcli_kwargs = dict({'config': config}, **toolcli_kwargs)

    toolcli.run_cli(
        raw_command=raw_command,
        command_index=command_index,
        **toolcli_kwargs,
    )


#
# # old
#


def create_ctc_cli():

    # hardcode this for speed
    # command_indices = {}
    # roots = [
    #     'ctc.cli.commands',
    #     'ctc.protocols.fei_utils.cli',
    # ]
    # for root in roots:
    #     branch = toolcli.autoindex_command_root(root_module_name=root)
    #     overlap = set(branch.keys()).intersection(command_indices.keys())
    #     if len(overlap) > 0:
    #         raise Exception('command_index key collision: ' + str(overlap))
    #     command_indices.update(branch)
    # print('command_indices:', command_indices)
    command_indices = {
        'ctc.cli.commands': [
            ('block',),
            ('gas',),
            ('transaction',),
            ('address',),
            ('token',),
            ('call',),
        ],
        'ctc.protocols.fei_utils.cli': [('fei', 'eth')],
        'ctc.protocols.chainlink_utils.cli': [('chainlink',)],
        'ctc.protocols.rari_utils.cli': [('rari', 'fuse')],
    }

    cli = toolcli.BaseCLI(
        command_indices=command_indices,
        passthrough_command=passthrough_command,
    )
    return cli


def passthrough_command(command):

    arg = command[1]

    if len(arg) <= 9 and str.isnumeric(arg):
        return 'block'

    elif arg.startswith('0x') and len(arg) == 66:
        return 'transaction'

    elif arg.startswith('0x') and len(arg) == 42:
        return 'address'

    elif str.isalnum(arg) and len(arg) <= 8:
        return 'token'

    else:
        raise Exception('could not determine command')




from __future__ import annotations

import typing

from ctc import directory
from ctc import rpc
from ctc import spec
from .. import address_utils
from .. import evm_spec


def get_erc20_address(token: spec.ERC20Reference) -> spec.ERC20Address:
    """return address of input token, input as either symbol or address"""
    if address_utils.is_address_str(token):
        return token
    elif isinstance(token, str):
        return directory.get_erc20_address(token)
    else:
        raise Exception('could not get token address')


#
# # generic erc20 calls
#


async def async_erc20_eth_call(
    function_name: str,
    token: spec.ERC20Reference,
    block: spec.BlockNumberReference = 'latest',
    **rpc_kwargs
) -> typing.Any:
    """perform eth_call for an erc20"""

    return await rpc.async_eth_call(
        to_address=get_erc20_address(token),
        function_abi=evm_spec.erc20_abis[function_name],
        block_number=block,
        **rpc_kwargs
    )


async def async_erc20s_eth_calls(
    function_name: str,
    tokens: typing.Iterable[spec.ERC20Reference],
    block: spec.BlockNumberReference = 'latest',
    **rpc_kwargs
) -> list[typing.Any]:
    """perform eth_call for multiple erc20s"""

    return await rpc.async_batch_eth_call(
        to_addresses=[get_erc20_address(token) for token in tokens],
        function_abi=evm_spec.erc20_abis[function_name],
        block_number=block,
        **rpc_kwargs
    )


async def async_erc20_eth_call_by_block(
    function_name: str,
    token: spec.ERC20Reference,
    blocks: typing.Iterable[spec.BlockNumberReference],
    **rpc_kwargs
) -> list[typing.Any]:
    """perform eth_call for an erc20 across multiple blocks"""

    return await rpc.async_batch_eth_call(
        to_address=get_erc20_address(token),
        function_abi=evm_spec.erc20_abis[function_name],
        block_numbers=blocks,
        **rpc_kwargs
    )




import typing

from ctc import directory
from ctc import rpc
from ctc import spec
from . import abi_utils


eip_897_abi = [
    {
        'name': 'proxyType',
        'type': 'function',
        'inputs': [],
        'outputs': [
            {'name': 'proxyTypeId', 'type': 'uint256'},
        ],
    },
    {
        'name': 'implementation',
        'type': 'function',
        'inputs': [],
        'outputs': [
            {'name': 'codeAddr', 'type': 'address'},
        ],
    },
]


async def async_get_eip897_proxy_type(
    contract_address: spec.Address, provider: spec.ProviderSpec = None
) -> int:

    function_abi = eip_897_abi[0]
    assert function_abi['name'] == 'proxyType'

    return await rpc.async_eth_call(
        to_address=contract_address,
        function_abi=function_abi,
        provider=provider,
    )


async def async_get_eip897_implementation(
    contract_address: spec.Address, provider: spec.ProviderSpec = None
) -> spec.Address:

    function_abi = eip_897_abi[1]
    assert function_abi['name'] == 'implementation'

    return await rpc.async_eth_call(
        to_address=contract_address,
        function_abi=function_abi,
        provider=provider,
    )


async def async_get_eip1967_proxy_logic_address(
    contract_address: spec.Address,
    block: typing.Optional[spec.BlockNumberReference] = None,
    provider: spec.ProviderSpec = None,
) -> spec.Address:
    """get a contract's logic address

    storage position obtained as:
        bytes32(uint256(keccak256('eip1967.proxy.implementation')) - 1)

    see https://eips.ethereum.org/EIPS/eip-1967
    """

    position = (
        '0x360894a13ba1a3210667c828492db98dca3e2076cc3735a920a3ca505d382bbc'
    )

    result = await rpc.async_eth_get_storage_at(
        address=contract_address,
        position=position,
        block_number=block,
        provider=provider,
    )

    return '0x' + result[-40:]


async def async_save_eip897_abi(
    contract_address: spec.Address, provider: spec.ProviderSpec = None
) -> None:

    provider = rpc.get_provider(provider)
    eip897_address = await async_get_eip897_implementation(
        contract_address,
        provider=provider,
    )

    network = provider['network']
    if network is None:
        raise Exception('could not determine network')

    abi_utils.async_save_proxy_contract_abi_to_filesystem(
        contract_address=contract_address,
        proxy_implementation=eip897_address,
        network=network,
    )


async def async_get_eip1967_proxy_beacon_address(
    contract_address: spec.Address,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> spec.Address:
    """get a contract's logic address

    storage position obtained as:
        bytes32(uint256(keccak256('eip1967.proxy.beacon')) - 1)

    see https://eips.ethereum.org/EIPS/eip-1967
    """

    position = (
        '0xa3f0ad74e5423aebfd80d3ef4346578335a9a72aeaee59ff6cb3582b35133d50'
    )

    result = await rpc.async_eth_get_storage_at(
        address=contract_address,
        position=position,
        block_number=block,
    )

    return '0x' + result[-40:]


async def async_get_eip1967_proxy_admin_address(
    contract_address: spec.Address,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> spec.Address:
    """get a contract's logic address

    storage position obtained as:
        bytes32(uint256(keccak256('eip1967.proxy.admin')) - 1)

    see https://eips.ethereum.org/EIPS/eip-1967
    """

    position = (
        '0xa3f0ad74e5423aebfd80d3ef4346578335a9a72aeaee59ff6cb3582b35133d50'
    )

    result = await rpc.async_eth_get_storage_at(
        address=contract_address,
        position=position,
        block_number=block,
    )

    return '0x' + result[-40:]

